%==========================================================================
% lossyFLAC.m
%==========================================================================
% David Robinson, 2007
%==========================================================================
% This code is open source. I'll pick a licence later (need advice), but:
% Any changes to this MATLAB code must be shared freely
% Any changes to the algorithm must be shared freely
% Any other implementations / interpretations must be shared freely
% You are free to include any implementation in commercial and/or closed
% source code, but must give an acknowledgement in help/about or similar.

% No warrantee / guarantee. This is work in progress and not debugged.

% Things to add:
% [dither - done]
% [force dither LSB - done]
% [put 20log10 outside of min() - done]
% [put conv() outside of selecting fft bins to keep - done]
% [buffer hanning - done]
%
% ms checking
% ignore digital silence
% 32k
% ReplayGain
% work on .wav files in blocks
% make FFT > block selection robust for all block sizes - done
% look one before and after for short FFT - done (optional)
% Add 31/32 + retain 5 bits declipping option - done
% Lossless correction file
% Try psy based spreading
% Try psy based noise shaping
% Try multi-frequency bands! (wavelet?)
% optimise, inc:

%==========================================================================
% internal constants (i.e. no need to change them)
%==========================================================================
clear

noise_averages=1000; % length of dither noise sample.
inaudible=1E-10; % small number to add to audio to prevent log(0).
minimum_bits_to_keep=2; % hard limit of bits to keep.
minimum_bits_to_keep_when_declipping=5; % ensure no clip on round.

%==========================================================================
% Set the source path
%==========================================================================
source_path=['C:\Data_Nic\_WAV\WAV\_WAV\'];
output_path=['C:\Data_Nic\_WAV\WAV\'];

%source_path=['Z:\My Music\Marillion - Real to Reel\'];
%source_path=['Z:\My Music\Mike Oldfield - The Songs of Distant Earth\'];
%output_path=['Z:\WAV\WAV\T1\'];

disp(['-----------------------------------------------------------------------------']);
disp(['Reading files from: ' source_path]);
disp(['Writing files to : ' output_path]);

process_start=now;

%==========================================================================
% external variables (i.e. 'switches' when implemented properly!)
%==========================================================================

noise_threshold_shift=0;
% noise threshold shift. average level of added quantisation
% noise relative to the lowest amplitude frequency bin (default=0=equal!)

low_frequency_limit=20;
high_frequency_limit=17226.5625;
% Frequency range over which to check for lowest amplitude signal

fix_clipped=3;
% 0 = do nothing
% 1 = reduce by 6dB and switch to 24-bit
% 2 = reduce by (2^(n-1)-1)/2^(n-1) (or (2^(n-1)-2)/2^(n-1) with triangular dither), stay at 16-bits, keep at least n bits
% 3 = Only reduce if there is a chance of clipping on rounding at maximum possible bits to remove (w.r.t. dither).

dither_choice=4;
% 0 = no dither
% 1 = rectangular dither
% 2 = triangular dither (+&- rand)
% 3 = gaussian dither using mean of gaussian_dither random numbers
gaussian_dither=6; % number of random numbers to mean for gaussian_dither
% 4 = triangular gaussian dither using mean of gaussian_dither random numbers
gaussian_triangular=6; % number of random numbers to mean for triangular_gaussian_dither

force_dither_LSB=1;
% 0 = no dither of LSB when no bits removed (default)
% 1 = force dither of LSB even when no bits removed (_theoretically_ "correct" when scaling audio for declipping)

btr_type=0;
% Use floor(mean(mean(bits_to_remove_table))) rather than min(min(bits_to_remove_table))
% in determining bits to remove for each codec_block. (now only 0 or 1);

codec_blocksize=1024;
% blocksize of lossless codec

process_ends_of_file=1;
% 0 = don't perform lossy calculation on ends of file unless all data is present to fill specified FFTs
% 1 = perform lossy calculation on ends of file even if some specified FFT data would lie beyond start or end of file

number_of_analyses=2;
switch number_of_analyses
case 2,
analysis_time=[1.451247e-3,2.321995e-2];
case 3,
analysis_time=[1.451247e-3,5.804989e-3,2.321995e-2];
case 4,
analysis_time=[1.451247e-3,5.804989e-3,2.321995e-2,9.287982e-2];
end

if (fix_clipped>=2),
minimum_bits_to_keep=max(minimum_bits_to_keep,minimum_bits_to_keep_when_declipping);
end;

%==========================================================================
% Get a list of all .wav files in that source path
%==========================================================================
filenamelist=dir([source_path '4*.wav']);
file_count=length(filenamelist);
disp(['Files to process: ' int2str(file_count)]);
disp(['-----------------------------------------------------------------------------']);

testing=0;

% Loop through them all, creating lossy versions
group_start=now;

for loop=1:length(filenamelist),
file_start=now;
source_name=filenamelist(loop).name;
param_string=[int2str(number_of_analyses) int2str(btr_type) int2str(dither_choice) int2str(fix_clipped) int2str(minimum_bits_to_keep) '.' ];
if testing==1,
output_name=[param_string source_name(1:length(source_name)-4) '.ss.wav'];
else
output_name=[source_name(1:length(source_name)-4) '.ss.wav'];
end

disp(['Loading ' source_name])

% load file
clear inaudio;

[inaudio,fs,bs]=wavreadraw([source_path source_name]);
[samples channels]=size(inaudio);
file_length=samples/fs;
maximum_bits_to_remove=bs-1-minimum_bits_to_keep;

peak_max=2^(bs-1)-1;
peak_min=-2^(bs-1);

peak_max_cnt=length(find(inaudio==peak_max));
peak_min_cnt=length(find(inaudio==peak_min));

amax=max(max(inaudio));
amin=min(min(inaudio));

disp(['Audio Peaks: ' int2str(amax) ', ' int2str(amin) '; Peaked=' int2str(peak_max_cnt+peak_min_cnt) '; Length: ' num2str(file_length,' %5.2f') ' seconds; Calculating........']);

clipping_reduction=1;
switch fix_clipped
case 1
% This pads it to 23 bits: It's treated as 24 bit data at the end, effectively dropping it by 6dB
inaudio = inaudio*(2^(23-bs));
bs=23;
force_dither_LSB=1;
case 2
if (dither_choice==2)||(dither_choice==4),
% Scale audio to peak 2 rounded LSBs away from digital full scale
clipping_reduction=((2^(minimum_bits_to_keep-1)-2)/2^(minimum_bits_to_keep-1));
else
% Scale audio to peak 1 rounded LSB away from digital full scale
clipping_reduction=((2^(minimum_bits_to_keep-1)-1)/2^(minimum_bits_to_keep-1));
end
case 3
% Scale audio to peak 1 rounded LSBs away from digital full
% scale (if necessary!)
clipping_reduction=min(1,(peak_max-2^(maximum_bits_to_remove-1))/max(amax,-amin));
end
if clipping_reduction<1,
disp(['Fix_Clipped; Reducing to ' num2str(clipping_reduction*100,'%5.2f') '% of original amplitude']);
inaudio=inaudio.*clipping_reduction;
force_dither_LSB=1;
end

% Add small value to audio data to avoid Log10(0)!
inaudio=inaudio+inaudible;

% Set up the FFT analysis lengths - define these in time (seconds)
% (will then use the nearest power of two based on sampling frequency)
clear spreading_function fft_length low_frequency_bin high_frequency_bin
clear reference_thresholds reference_threshold min_bin bits_to_remove bits_to_remove_table

for analysis_count=1:number_of_analyses,
spreading_function{analysis_count}=[0.25,0.25,0.25,0.25];
fft_overlap(analysis_count)=2;
end_overlap(analysis_count)=0.5;
end %for analysis_count

% spreading function to apply to FFT before determining lowest amplitude
% Keep peak at centre, even if it means padding with zeros
% FFT overlap - number of FFTs to start within 1 FFT length.
% 1 = no overlap; 2 = 50% overlap; 4 = 75% overlap
% Maximum amount of FFT to overlap over edge of codec block before that
% FFT is ignored within that block (default = 0.5)

% Loop through each analysis length (typically only two) and set the FFT values
for analysis_number=1:number_of_analyses,
% Calculate the closest FFT length
fft_length(analysis_number)=2^round(log10(analysis_time(analysis_number)*fs)/log10(2));

% Generate window function
window_function{analysis_number}=hanning(fft_length(analysis_number));

% Calculate which FFT bin corresponds to the low frequency limit
low_frequency_bin(analysis_number)=round(fft_length(analysis_number)*low_frequency_limit/fs);
if low_frequency_bin(analysis_number)<2, low_frequency_bin(analysis_number)=2; end
if low_frequency_bin(analysis_number)>fft_length(analysis_number)/2, error('low frequency too high'); end

% Calculate which FFT bin corresponds to the high frequency limit
high_frequency_bin(analysis_number)=round(fft_length(analysis_number)*high_frequency_limit/fs);
if high_frequency_bin(analysis_number)<2, error('high frequency too low'); end
if high_frequency_bin(analysis_number)>fft_length(analysis_number)/2, high_frequency_bin(analysis_number)=fft_length(analysis_number)/2; end

% Calculate FFT bins to ignore due to convolution
conv_bin1(analysis_number)=1+round((length(spreading_function{analysis_number})-1)/2);
conv_bin2(analysis_number)=(2+high_frequency_bin(analysis_number)-low_frequency_bin(analysis_number))-conv_bin1(analysis_number);

buffered_min_bin_start(analysis_number)=0;
end
variables_filename=['lossy_variables__cbs' num2str(codec_blocksize) '_fs' num2str(fs) '_bs' num2str(bs) '_noa' num2str(number_of_analyses) '_fft' num2str(fft_length) '_lfb' num2str(low_frequency_bin) '_hfb' num2str(high_frequency_bin) '_nts' num2str(noise_threshold_shift) '_dc' num2str(dither_choice) '.mat'];

% Find out if we've stored the quantisation noise thresholds before
if exist(variables_filename,'file'),
load(variables_filename)
% If not, estimate quantisation noise at each bit in these FFTs
else
disp('Thresholds not calculated for these parameters - calculating...')
for analysis_number=1:number_of_analyses,
clear reference_thresholds
reference_thresholds(1:noise_averages,1:bs)=zeros;
for av_number=1:noise_averages,
clear dither_sample
% Create a simulated input signal from noise
fftl=fft_length(analysis_number);
signal_sample=floor((rand(fftl,1)-0.5).*((2^bs)));
switch dither_choice,
case 0, % no dither
dither_sample(1:fftl,1)=zeros;
case 1, % rectangular dither - not recommended.
dither_sample=(rand(fftl,1)-0.5);
case 2, % triangular dither
dither_sample=(rand(fftl,1)-rand(fftl,1));
case 3, % gaussian dither
dither_sample=(mean(rand(fftl,gaussian_dither),2)-0.5);
case 4, % triangular gaussian dither?
dither_sample=(mean(rand(fftl,gaussian_triangular),2)-mean(rand(fftl,gaussian_triangular),2));
otherwise error('unknown dither choice');
end

for bits_to_remove=1:bs,
% This models the quantisation noise introduced by removing the last "bits_to_remove" bits from the audio data:
this_noise_sample=(round((signal_sample/(2^bits_to_remove))+dither_sample).*2^bits_to_remove)-signal_sample;
fft_result=abs(fft(this_noise_sample.*window_function{analysis_number}));
fft_result2=conv(fft_result(low_frequency_bin(analysis_number):high_frequency_bin(analysis_number)),spreading_function{analysis_number});
reference_thresholds(av_number,bits_to_remove)=mean(20*log10(fft_result2(conv_bin1(analysis_number):conv_bin2(analysis_number))));
end
end
reference_threshold{analysis_number}=mean(reference_thresholds)-noise_threshold_shift;

for threshold=1:round(20*log10(2^(bs+4))),
if isempty(find(reference_threshold{analysis_number}<threshold)),
threshold_index{analysis_number}(threshold)=0;
else
threshold_index{analysis_number}(threshold)=max(find(reference_threshold{analysis_number}<threshold));
end
end
end
save('-MAT',variables_filename,'threshold_index')
disp('Calculation complete - Thresholds stored')
end

clear bits_to_remove
bits_to_remove(1:ceil(samples/codec_blocksize))=zeros;

if process_ends_of_file==0,
start_pad=ceil(max(end_overlap.*fft_length)/codec_blocksize)*codec_blocksize;
end_pad=codec_blocksize+max(end_overlap.*fft_length)-1;
else
start_pad=0;
end_pad=0;
end

% loop through codec blocks
for codec_block_start=1+start_pad:codec_blocksize:samples-end_pad,
codec_block_number=1+round(codec_block_start/codec_blocksize);

codec_block_end=codec_block_start+codec_blocksize-1;
codec_block_middle=codec_block_start+floor(codec_blocksize/2);

if codec_block_end>samples, codec_block_end=samples; end % Don't jump past end of file!

for analysis_number=1:number_of_analyses,

analysis_blocks_start=codec_block_start-round(end_overlap(analysis_number)*fft_length(analysis_number));
analysis_blocks_end=codec_block_end+round(end_overlap(analysis_number)*fft_length(analysis_number));

central_block_start=codec_block_middle-round(fft_length(analysis_number)/2);

half_number_of_blocks=floor((central_block_start-analysis_blocks_start)/round(fft_length(analysis_number)/fft_overlap(analysis_number)));

actual_analysis_blocks_start=central_block_start-half_number_of_blocks*round(fft_length(analysis_number)/fft_overlap(analysis_number));
actual_analysis_blocks_end=central_block_start+(half_number_of_blocks+2)*round(fft_length(analysis_number)/fft_overlap(analysis_number))-1;

for analysis_block_start=actual_analysis_blocks_start:round(fft_length(analysis_number)/fft_overlap(analysis_number)):actual_analysis_blocks_end-round(fft_length(analysis_number)/fft_overlap(analysis_number)),
analysis_block_number=1+((analysis_block_start-actual_analysis_blocks_start)/round(fft_length(analysis_number)/fft_overlap(analysis_number)));

% If we're not processing the ends of the file, then the
% calculated FFT analysis block boundaries are correct
if process_ends_of_file==0,
actual_analysis_block_start=analysis_block_start;
else
% If we are processing the ends of the file, then make
% sure we don't attempt to look before the start or
% after the end of the file
actual_analysis_block_start=max(analysis_block_start,1);
actual_analysis_block_start=min(actual_analysis_block_start,1+samples-fft_length(analysis_number));
end

% If we've buffered these values before (due to overlapping blocks), use them
% if buffered_min_bin_start(analysis_number)==analysis_block_start,
% min_bin{analysis_number}(analysis_block_number,1:channels)=buffered_min_bin{analysis_number}(1,1:channels);
% else % otherwise calculate them
for channel=1:channels,
fft_result=abs(fft(window_function{analysis_number}.*inaudio(actual_analysis_block_start:actual_analysis_block_start+fft_length(analysis_number)-1,channel)));
fft_result2=conv(fft_result(low_frequency_bin(analysis_number):high_frequency_bin(analysis_number)),spreading_function{analysis_number});
min_bin{analysis_number}(analysis_block_number,channel)=20*log10(min(fft_result2(conv_bin1(analysis_number):conv_bin2(analysis_number))));
end
%end
end

% Store the last result in a buffer in case we can use it next time due to block overlap (saves time)
%buffered_min_bin{analysis_number}(1,1:channels)=min_bin{analysis_number}(analysis_block_number,1:channels);
%buffered_min_bin_start(analysis_number)=actual_analysis_block_start;

for channel=1:channels,
this_min_bin=round(min(min_bin{analysis_number}(:,channel)));
if this_min_bin<1, % i.e. if it's quieter than quantisation noise at the least significant bit
bits_to_remove_table(analysis_number,channel)=0; % don't remove any bits!
else
bits_to_remove_table(analysis_number,channel)=min(threshold_index{analysis_number}(this_min_bin),maximum_bits_to_remove);
end
end
end

if (btr_type==0),
bits_to_remove(codec_block_number)=min(min(bits_to_remove_table));
else
bits_to_remove(codec_block_number)=floor(mean(mean(bits_to_remove_table.^2))^0.5);
end;

tvl=2^bits_to_remove(codec_block_number);
tbl=1+codec_block_end-codec_block_start;
cbs=codec_block_start;
cbe=codec_block_end;
chs=channels;

if bits_to_remove(codec_block_number)>0 | force_dither_LSB==1,
switch dither_choice,
case 0, % no dither
inaudio(cbs:cbe,1:chs)=round(inaudio(cbs:cbe,1:chs)/tvl).*tvl;
case 1, % rectangular dither
inaudio(cbs:cbe,1:chs)=round(inaudio(cbs:cbe,1:chs)/tvl+rand(tbl,chs)-0.5).*tvl;
case 2, % triangular dither
inaudio(cbs:cbe,1:chs)=round(inaudio(cbs:cbe,1:chs)/tvl+(rand(tbl,chs)-rand(tbl,chs))).*tvl;
case 3, % gaussian dither
inaudio(cbs:cbe,1:chs)=round(inaudio(cbs:cbe,1:chs)/tvl+(mean(rand(tbl,chs,gaussian_dither),3)-0.5)).*tvl;
case 4, % triangular gaussian dither?
inaudio(cbs:cbe,1:chs)=round(inaudio(cbs:cbe,1:chs)/tvl+(mean(rand(tbl,chs,gaussian_triangular),3)-mean(rand(tbl,chs,gaussian_triangular),3))).*tvl;
end

rounded_bit_removed_peak=(2^(bs-1))-tvl;
block_peak_max=max(max(inaudio(cbs:cbe,1:chs)));
if (block_peak_max>peak_max),
peaks_over=length(find(inaudio(cbs:cbe,1:chs)>peak_max));
inaudio(cbs:cbe,1:chs)=min(inaudio(cbs:cbe,1:chs),rounded_bit_removed_peak);
disp(['Clipped Peak(s) Over at Block Number: ' int2str(codec_block_number) ' for ' int2str(peaks_over) ' samples. Max= ' int2str(block_peak_max)]);
end;
block_peak_min=min(min(inaudio(cbs:cbe,1:chs)));
if (block_peak_min<peak_min),
peaks_under=length(find(inaudio(cbs:cbe,1:chs)<peak_min));
inaudio(cbs:cbe,1:chs)=max(inaudio(cbs:cbe,1:chs),-rounded_bit_removed_peak);
disp(['Clipped Peak(s) Under at Block Number: ' int2str(codec_block_number) ' for ' int2str(peaks_under) ' samples. Min= ' int2str(block_peak_min)]);
end
end
end

if fix_clipped==1, bs=24; end

amax_cnt=length(find(inaudio>peak_max));
amin_cnt=length(find(inaudio<peak_min));
peak_max_cnt=length(find(inaudio==peak_max));
peak_min_cnt=length(find(inaudio==peak_min));

disp(['Audio Peaks: ' int2str(max(max(inaudio))) ', ' int2str(min(min(inaudio))) '; Clipped=' int2str(amax_cnt+amin_cnt) '; Peaked=' int2str(peak_max_cnt+peak_min_cnt) '; Average BTR:' num2str(mean(bits_to_remove),'% 5.2f') '; Time:' num2str((now-file_start)*24*3600,'% 5.2f') ' Rate: ' num2str(file_length/((now-file_start)*24*36),'%3.0f') '%']);

for btrs_i=0:15,
btr_summary(btrs_i+1)=length(find(bits_to_remove==btrs_i))/length(bits_to_remove);
end
disp(['BTR: ' num2str(btr_summary(1:16).*100,' %4.1f')]);

disp(['Saving ' output_name]);
disp([' ']);

wavwriteraw(inaudio,fs,bs,[output_path output_name])

% Make a .bat file to call FLAC twice for comparison: lossless and lossy
% Note comparison might not be fair, since -b1024 itself gives better
% compression on _some_ samples
% fid=fopen('temp.bat','w');
% dummy=fprintf(fid,'%s\n',['"C:\Program Files\FLAC\flac.exe" "' source_path filename '"']);
% dummy=fprintf(fid,'%s\n',['"C:\Program Files\FLAC\flac.exe" -b' num2str(codec_blocksize) ' "' source_path filename(1:length(filename)-4) '.lossy.wav"']);
% dummy=fclose(fid);
% Run the .bat file
% !temp.bat

%==========================================================================
end % loop through files
%==========================================================================

%==========================================================================
disp(['-----------------------------------------------------------------------------']);
disp(['Elapsed time: ' num2str((now-group_start)*24*3600,'%6.2f')])
disp(['-----------------------------------------------------------------------------']);

%==========================================================================
disp(['Total elapsed time: ' num2str((now-process_start)*24*3600,'%6.2f')])
disp(['-----------------------------------------------------------------------------']);
%==========================================================================
